/**
 * MPQAudioService.java: Service to control audio recording and playback.
 *            Copyright (C) 2012 donfede@casagrau.org
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * created 2012-11-14 donfede@casagrau.org
 * credit to: android API Guide on Services -- (Apache2) http://developer.android.com/guide/components/services.html , http://developer.android.com/guide/components/bound-services.html
 *			  Reto Meier, EMEA Android Developer Advocate, Google UK; for a record and playback example -- http://eurodev.blogspot.com/2009/09/raw-audio-manipulation-in-android.html
 *			  Stjepan Rajko, urbanSTEW -- (GPL) RehearsalAssistant/RehearsalAudioRecorder.java
 *			  cristi (Cristian Baita) -- http://indyvision.net/2010/02/android-threads-tutorial-part-3/
 *			  davidkuo -- http://codinghard.wordpress.com/2009/05/16/android-thread-messaging/
 * $Id: MPQAudioService.java,v 1.2 2012-12-19 23:15:10 donfede Exp $
 */

package org.casagrau.mpq.audio_service;

import java.io.File;
import java.io.RandomAccessFile;

import org.casagrau.mpq.mpq_android.MPQMainTabActivity;
import org.casagrau.mpq.mpq_android.R;
import org.casagrau.mpq.mpq_android.SessionsDBProvider.SessionsDB;

import android.annotation.SuppressLint;
import android.app.Notification;
import android.app.PendingIntent;
import android.app.Service;
import android.content.Context;
import android.content.Intent;
import android.media.AudioFormat;
import android.media.AudioManager;
import android.os.Binder;
import android.os.Bundle;
import android.os.Environment;
import android.os.Handler;
import android.os.IBinder;
import android.os.Message;
import android.support.v4.app.NotificationCompat;
import android.util.Log;

@SuppressLint("HandlerLeak")
public class MPQAudioService extends Service {
	private static final String LOGTAG = "MPQAudioService";
	
	public static final String ACTION_REC = "org.casagrau.mpq.audio_service.intent.REC";				// Intent Action to begin recording a file; Requires a filename to create (or overwrite).
	public static final String ACTION_PAUSE_PLAY = "org.casagrau.mpq.audio_service.intent.PAUSE_PLAY";// Intent Action to pause/un-pause playing.
	public static final String ACTION_PAUSE_REC = "org.casagrau.mpq.audio_service.intent.PAUSE_REC";	// Intent Action to pause/un-pause recording.
	public static final String ACTION_STOP = "org.casagrau.mpq.audio_service.intent.STOP";			// Intent Action to stop playing or recording.
	public static final String ACTION_PLAY = "org.casagrau.mpq.audio_service.intent.PLAY";			// Intent Action to begin playing a file; Either requires a filename to play, or a previous called to open or play a file.
	public static final String ACTION_OPEN = "org.casagrau.mpq.audio_service.intent.OPEN";			// Intent Action to open a file (for playing later); Requires a filename to a file that may be played.
	public static final String ACTION_BIND = "org.casagrau.mpq.audio_service.intent.BIND";			// Intent Action to bind to the service.
	// set cursor pos "ACTION_SET_CURSOR" "ACTION_SET_CURSOR_POS" FIXME: implement, and figure what to call this

	// example/RecForge notifications/states: Ready, Recording, Paused_Play, Paused_Rec, Stop, Play, Init 
	public static final int FLG_SRV_STATE_INIT = -1;		// Service is initializing.
	public static final int FLG_SRV_STATE_READY = 0;		// Service is ready.
	public static final int FLG_SRV_STATE_REC = 1;			// Service is recording.
	public static final int FLG_SRV_STATE_PAUSE_PLAY = 2;	// Service is paused from playing.
	public static final int FLG_SRV_STATE_PAUSE_REC = 3;	// Service is paused from recording.
	public static final int FLG_SRV_STATE_STOP = 4;			// Service is stopping.
	public static final int FLG_SRV_STATE_PLAY = 5;			// Service is playing.
	
	public static final String ARG_FILENAME = "org.casagrau.mpq.audio_service.audio_service_filename"; 	// Argument handle for filename.
	public static final String ARG_FILEPOS  = "org.casagrau.mpq.audio_service.audio_service_fileposition";// Argument handle for cursor position in file, in milliseconds.
	public static final String DEFAULT_FILENAME = "/mpq/testAudioService01.wav";	// Default filename to use. (yes, a wart, but a 'contained' one)
	// FIXME: currently have a hard coded Activity class to run after a notification -- ASMainActivity.class;
	
	public static final String ARG_PMSG_STATE = "state"; 		// Argument for parent messages with new state.
	public static final String ARG_PMSG_FILEPOS = "filepos"; 	// Argument for parent messages with new cursor position in milliseconds.
	public static final String ARG_PMSG_FILELEN  = "filelength";// Argument for parent messages with new *Audio* file length, in milliseconds.
	
	private static final int NOTIFICATION_ID = R.string.notification_aservice; // FIXME: review this constant name -- too close to "NOTIFICATION_SERVICE"

	
	private int mState = FLG_SRV_STATE_INIT;		// State of the audio service.  See FLG_SRV_STATE_INIT and related constants for options.
	private String mFilename;						// Filename of the file being worked on by the audio service.
	private long mFilepos;							// File position of the cursor, in milliseconds.
	private long mFilelen;							// File *Audio* length, in milliseconds.
	private int mVolumeLevel;						// Latest volume level (from mic or speaker).
	private MPQAudioServiceListener mStateListener;	// Handle to a state change listener in the associated activity.
	private IBinder mBinder;						// Copy of the binder interface to hand out.
	private AudioManager mAM;						// Object handle to the audio manager, used to get and release focus.
	private MPQAudioWorkerThread mWorkerThread;		// Handle to worker thread	
	private Handler mParentHandler;					// Parent message handler (this), to receive messages from child threads.
    
    
	//////////////
	// Callback listener interface
	/**
	 * Define callback interface to send messages back to the calling activity.
	 * FIXME: this could likely be replaced with the android message/handler system at some point (or broadcast/receiver); but, it is working fine for now.
	 */
	public interface MPQAudioServiceListener {
		public void onMPQAudioServiceStateChange(int newState);
		public void onMPQAudioServiceAudioChange(long cursorPos, long fileLen, int volumeLevel);
	}
	
    
	/*****************************************
     * Overridden Service methods
     */

	@Override
    public void onCreate() {
        Log.v(LOGTAG, "onCreate() 00 service created");
        mAM = (AudioManager) getSystemService(Context.AUDIO_SERVICE);
        
    	mFilename = null;
    	mFilepos = -1;
    	mFilelen = -1;
    	mVolumeLevel = -1;
        
        // Service message Handler, receiving messages from child worker threads.
        mParentHandler = new Handler() {	
        	public void handleMessage(Message msg) {
        		Log.i(LOGTAG, "mParentHndler handleMessage() msg: " + msg + " obj: " + (String)msg.obj);

        		Bundle bundle = msg.getData();
        		if (null != bundle) {
        			int lstate = bundle.getInt(ARG_PMSG_STATE, -1);
        			Log.i(LOGTAG, "mParentHandler handleMessage() 12 new lstate: " + lstate + " old mState: " + mState);
        			if ( (lstate >= 0) && (lstate != mState) ) {
        				mState = lstate;
        				triggerStateListener();
        			}

        			boolean lupdateFlag = false;
        			long laudioPos = bundle.getLong(ARG_PMSG_FILEPOS, -1);
        			Log.i(LOGTAG, "mParentHandler handleMessage() 14 new audioPos: " + laudioPos + " old audioPos: " + mFilepos);
        			if ( (laudioPos >= 0) && (laudioPos != mFilepos) ) {
        				mFilepos = laudioPos;
        				if (FLG_SRV_STATE_REC == mState) // If recording, update the file length to the current recording position.  A bit warty to have this here... but the alternative is to make 2x calls from worker, or make a new method to send Pos + Len together.
        					mFilelen = mFilepos;
            			lupdateFlag = true;
        			}
           			long lfileLen = bundle.getLong(ARG_PMSG_FILELEN, -1);
        			Log.i(LOGTAG, "mParentHandler handleMessage() 16 new filelen: " + lfileLen+ " old filelen: " + mFilelen);
        			if ( (lfileLen >= 0) && (lfileLen != mFilelen) ) {
        				mFilelen = lfileLen;
            			lupdateFlag = true;
        			}
    				if (true == lupdateFlag)
    					triggerAudioListener();
    				
        		} else Log.e(LOGTAG, "mParentHandler handleMessage() 13 no bundle");
        	}
        }; // END mParentHandler Handler()
        
    } // END onCreate()
    
    /**
     * The service is starting, due to a call to startService().
     * Handle one of the valid option type for the Audio Service.
     * @returns int with code for START_CONTINUATION_MASK
	 * @throws IllegalArgumentException if an invalid Intent action is received, 
	 * 									or if getExtras() arguments are missing.
	 * 
     * This method aligns with, and enforces, the MPQAudioService design and state diagram.
     */
    @Override
    public int onStartCommand(Intent intent, int flags, int startId) {
        Log.v(LOGTAG, "onStartCommand() 00 intent: " + intent + " flags: " + flags + " start id: " + startId);

        String lAction = intent.getAction();
        Log.d(LOGTAG, "onStartCommand() 10 intent action: " + lAction);        
        
        // Parse the Intent action, check permissions, and do the appropriate activity.
        if (lAction.equals(ACTION_REC)) {
        	// Valid state check
        	if ( (FLG_SRV_STATE_INIT != mState) &&
        			(FLG_SRV_STATE_READY != mState) ) {
            	throw new IllegalArgumentException("onStartCommand() illegal permission.  state: " + mState + " lIAction: " + lAction);
        	}
        	getAudioFocus();
            showForegroundNotification(R.string.notification_aservice_rec);
			// set state to recording
        	mState = FLG_SRV_STATE_REC;
        	mFilepos = 0;
        	mFilelen = 0;
        	// Get file from Intent "extras".
    		Bundle lextraArgs = intent.getExtras();
    		String lFilename = null;
    		long lFilepos = 0;
    		if (null != lextraArgs) {
    			lFilename = lextraArgs.getString(MPQAudioService.ARG_FILENAME);
    			if (lFilename != null) mFilename = lFilename;
    			lFilepos = lextraArgs.getLong(MPQAudioService.ARG_FILEPOS, -1);
    			if (lFilepos >= 0) mFilepos = lFilepos;
    		} else throw new IllegalArgumentException("onStartCommand() REC missing filename.  state: " + mState + " lIAction: " + lAction);
        	// Start worker thread to record "indefinitely".
    		mWorkerThread = new MPQAudioWorkerRecThread(mParentHandler, mFilename, mFilepos, mState, (Context) this);
            mWorkerThread.start();        		
        } else if (lAction.equals(ACTION_PAUSE_REC)) {
        	// Valid state check
        	if ( (FLG_SRV_STATE_PAUSE_REC != mState) &&
        			(FLG_SRV_STATE_REC != mState) ) {
            	throw new IllegalArgumentException("onStartCommand() illegal permission.  state: " + mState + " lIAction: " + lAction);
        	}
        	// Toggle of pause button, while recording.
        	if (FLG_SRV_STATE_PAUSE_REC == mState) {        		
        		getAudioFocus();
	            showForegroundNotification(R.string.notification_aservice_rec);
                // set state to resume recording
        		mState = FLG_SRV_STATE_REC;
        	} else {
        		releaseAudioFocus();
	            showForegroundNotification(R.string.notification_aservice_pause_rec);
            	// set state to paused
        		mState = FLG_SRV_STATE_PAUSE_REC;
        	}
        	// send new state to child worker thread 
        	sendCMsgState(mState);
        } else if (lAction.equals(ACTION_PAUSE_PLAY)) {
        	// Valid state check
        	if ( (FLG_SRV_STATE_PAUSE_PLAY != mState) &&
        			(FLG_SRV_STATE_PLAY != mState) ) {
            	throw new IllegalArgumentException("onStartCommand() illegal permission.  state: " + mState + " lIAction: " + lAction);
        	}
        	// Toggle of pause button, while playing.
        	if (FLG_SRV_STATE_PAUSE_PLAY == mState) {
        		getAudioFocus();
	            showForegroundNotification(R.string.notification_aservice_play);
        		// set state to resume playing
        		mState = FLG_SRV_STATE_PLAY;
        	} else {
        		releaseAudioFocus();
	            showForegroundNotification(R.string.notification_aservice_pause_play);
	        	// set state to paused
            	mState = FLG_SRV_STATE_PAUSE_PLAY;
        	}
        	// send new state to child worker thread 
        	sendCMsgState(mState);
        } else if (lAction.equals(ACTION_PLAY)) {
        	// Valid state check
        	if ( (FLG_SRV_STATE_INIT != mState) &&
        			(FLG_SRV_STATE_READY != mState)	) {
            	throw new IllegalArgumentException("onStartCommand() illegal permission.  state: " + mState + " lIAction: " + lAction);
        	}
        	// get file from intent "extra" (optional, provided a file has been pre-opened/played/recorded)
        	getAudioFocus();
            showForegroundNotification(R.string.notification_aservice_play);
        	// set state to playing
        	mState = FLG_SRV_STATE_PLAY;
        	// Get file from Intent "extras".
    		Bundle lextraArgs = intent.getExtras();
    		String lFilename = null;
    		long lFilepos = 0;
    		if (null != lextraArgs) {
    			lFilename = lextraArgs.getString(MPQAudioService.ARG_FILENAME);
    			lFilepos = lextraArgs.getLong(MPQAudioService.ARG_FILEPOS, -1);
    		} 
        	if (null == mWorkerThread) {
    			if (lFilename != null) mFilename = lFilename;
    			if (lFilepos >= 0) mFilepos = lFilepos;
            	mFilelen = -1; // Reset audio length, to be updated by the playback thread in moments.
    			// Start worker thread to play until end of file.
    			mWorkerThread = new MPQAudioWorkerPlayThread(mParentHandler, mFilename, mFilepos, mState, (Context) this);
                mWorkerThread.start();
        	} else {
            	sendCMsgState(mState);
            	if (lFilepos >= 0) sendCMsgPos(lFilepos);
        	}
        } else if (lAction.equals(ACTION_STOP)) {
        	// Valid state check
        	if ( (FLG_SRV_STATE_PLAY != mState) &&
        			(FLG_SRV_STATE_PAUSE_PLAY != mState) &&
        			(FLG_SRV_STATE_REC != mState) &&
        			(FLG_SRV_STATE_PAUSE_REC != mState) &&
        			(FLG_SRV_STATE_STOP != mState) ) {
            	throw new IllegalArgumentException("onStartCommand() illegal permission.  state: " + mState + " lIAction: " + lAction);
        	}
        	// stop notification and stop foreground service
        	stopForeground(true);
        	releaseAudioFocus();
        	// set state to stop 
        	mState = FLG_SRV_STATE_STOP;
        	mFilename = null;
        	mFilepos = -1;
        	mFilelen = -1;
        	mVolumeLevel = -1;
        	sendCMsgState(mState);
        	// Release the worker thread.
    		mWorkerThread = null;
        } else if (lAction.equals(ACTION_OPEN)) {
// FIXME: evil hack - to get around open crashing if we're recording.  proper fix is to bind from SessionDetailFragment...        	
if (FLG_SRV_STATE_REC == mState) return START_NOT_STICKY;        	
        	// Valid state check        	
        	if ( (FLG_SRV_STATE_INIT != mState) &&
        			(FLG_SRV_STATE_STOP != mState) &&
        			(FLG_SRV_STATE_READY != mState) ) {
            	throw new IllegalArgumentException("onStartCommand() illegal permission.  state: " + mState + " lIAction: " + lAction);
        	}
        	// Get file from Intent "extras".
    		Bundle lextraArgs = intent.getExtras();
    		String lFilename = null;
    		if (null != lextraArgs) {
    			lFilename = lextraArgs.getString(MPQAudioService.ARG_FILENAME);
    		} else throw new IllegalArgumentException("onStartCommand() OPEN missing filename.  state: " + mState + " lIAction: " + lAction);
        	
        	// Set state to ready if there is a valid file with the Intent.
			if (lFilename != null) {
				mFilename = lFilename;
				mFilepos = 0;
	        	mFilelen = -1;
	            showForegroundNotification(R.string.notification_aservice_ready);
	        	mState = FLG_SRV_STATE_READY;
			}
        } else {
        	// Raise exception
        	Log.d(LOGTAG, "onStartCommand() 98 RAISING EXCEPTION");
        	throw new IllegalArgumentException("onStartCommand() lIAction: " + lAction);
        }

        // Notify listener of state change.
		triggerStateListener();
		
        return START_STICKY; // see START_CONTINUATION_MASK for other options to return, such as START_NOT_STICKY or START_REDELIVER_INTENT
    } // END onStartCommand()
    
     
    
    //////////////////////////////////
    //
    // Binder interface for this service.
    /**
     * Binder class for the MPQ Audio Service.
     */
    public class LocalBinder extends Binder {
        /**
         * Return a handle to this instance of the Audio Service service, so clients can call public methods
         * @return MPQAudioService
         */
    	public MPQAudioService getService() {
            return MPQAudioService.this;
        }
    } // END class LocalBinder

    /**
     * Method called the first time a client binds to the service.
     * @param intent to bind to the service
     * @return IBinder 
     */
	@Override
	public IBinder onBind(Intent intent) {
        Log.d(LOGTAG, "onBind() intent: " + intent);
    	mBinder = new LocalBinder();
        return mBinder;
	}
	
	
	

    
    @Override
    public void onDestroy() {
        Log.d(LOGTAG, "onDestroy()");
        // The service is no longer used and is being destroyed
        releaseAudioFocus(); // just in case
    }

    
    //////////////////////////////////
    //
    // public methods for the class, used by the binder
    
    
    /**
     * Return the state of the Audio Service via the binder.
     * @return int with a code identifying the audio service state.
     */
    public int getStateBnd() {
 		Log.d(LOGTAG, "getStateBnd() 00 state: " + mState);
    	return mState;
    }
    
    /**
     * Return the current file, via the binder, if the state is REC, PLAY, PAUSE, or READY.
     * @return String with current filename if a valid state, otherwise return null.
     */
    public String getFilenameBnd() {
 		Log.d(LOGTAG, "getFilenameBnd() 00 mFilename: " + mFilename);
    	String lFilename = null;
		if ( (FLG_SRV_STATE_REC == mState) ||
				(FLG_SRV_STATE_PLAY == mState) ||
				(FLG_SRV_STATE_PAUSE_REC == mState) ||
				(FLG_SRV_STATE_PAUSE_PLAY == mState) ||
				(FLG_SRV_STATE_READY == mState) ) {
			lFilename = mFilename;
		}
    	return lFilename;
    } // END getFilenameBnd()
    
    /**
     * Return the current cursor position in the audio file (excluding headers), via the binder.
     *   State must be REC, PLAY, PAUSE, or READY.
     * @return long with current audio file position in milliseconds, or -1 if there is an error.
     */
    public long getFileposBnd() {
    	Log.d(LOGTAG, "getFileposBnd() 00 mFilename: " + mFilename + " mFilepos: " + mFilepos);
    	long lFilepos = -1;
    	if ( (FLG_SRV_STATE_REC == mState) ||
				(FLG_SRV_STATE_PLAY == mState) ||
				(FLG_SRV_STATE_PAUSE_REC == mState) ||
				(FLG_SRV_STATE_PAUSE_PLAY == mState) ||
				(FLG_SRV_STATE_READY == mState) ) {
			lFilepos = mFilepos;
		}
    	return lFilepos;
    } // END getFileposBnd()
    
    /**
     * Return the current volume level, via the binder.
     *   State must be REC, PLAY, PAUSE.
     * @return int with the latest amplitude.
     */
    public int getAudioVolBnd() {
    	Log.d(LOGTAG, "getAudioVolBnd() 00 mFilename: " + mFilename + " mFilepos: " + mFilepos + " mVolumeLevel: " + mVolumeLevel);
    	int lVolumeLevel = -1;
    	if ( (FLG_SRV_STATE_REC == mState) ||
				(FLG_SRV_STATE_PLAY == mState) ||
				(FLG_SRV_STATE_PAUSE_REC == mState) ||
				(FLG_SRV_STATE_PAUSE_PLAY == mState) ) {
			lVolumeLevel = mVolumeLevel;
		}
    	return lVolumeLevel;
    } // END getAudioVolBnd()
    
    
    /**
     * Return the current file length (ms), via the binder.
     * @return long with current file length in milliseconds or -1 if there is an error.
     */
    public long getFileLengthMSBnd() {
    	Log.d(LOGTAG, "getFileLengthMSBnd() 00  filename:" + mFilename + " old mFilelen: " + mFilelen);
    	//  Convert bytes to milliseconds, based on channelCount,frequency, and sampleBit. // FIXME: hard coded now, read from wave header later...
    	long lfileMS = -1;
    	
    	if (null != mFilename) {
    		if (mFilelen >= 0) {
    	    	Log.d(LOGTAG, "getFileLengthMSBnd() 09  CACHED len:" + mFilelen);
    			lfileMS = mFilelen;
    		} else {
    			try {
		    		// FIXME: this block is from MPQAudioWorkerPlayThread -- likely a foundation for WAVE reader method... (though no sharing between that scope and this ... maybe static?? ) 
					// Open file to play back
		    		File lFile = new File(Environment.getExternalStorageDirectory().getAbsolutePath() + mFilename);
		    	    
					RandomAccessFile lFileHandle = new RandomAccessFile(lFile, "r");
		
					// FIXME: read WAVE file headers
					int lfrequency = 11025; // cycles per second
					double lfrequencyMS = lfrequency / 1000.0; // cycles per millisecond
					int lchannelConfiguration = AudioFormat.CHANNEL_OUT_MONO;
					int laudioEncoding = AudioFormat.ENCODING_PCM_16BIT; 
					
			    	short lchannelCount;
			    	if (AudioFormat.CHANNEL_OUT_MONO == lchannelConfiguration) lchannelCount = 1;
			    	else if (AudioFormat.CHANNEL_OUT_STEREO == lchannelConfiguration) lchannelCount = 2;
			    	else throw new IllegalStateException("getFileLengthMSBnd() unsuported channel configuration: " + lchannelConfiguration);
		
			    	short lsampleBit; // sample bitrate
			    	if (AudioFormat.ENCODING_PCM_16BIT == laudioEncoding) lsampleBit = 16;
			    	else if (AudioFormat.ENCODING_PCM_8BIT == laudioEncoding) lsampleBit = 8;
			    	else throw new IllegalStateException("getFileLengthMSBnd() unsuported sample bitrate: " + lchannelConfiguration);
			    	
			    	int lframeSize = (lsampleBit/8)*lchannelCount;
			    	Log.d(LOGTAG, "getFileLengthMSBnd() 20  framesize:" + lframeSize + " channelCount: " + lchannelCount + " sampleBit: " + lsampleBit);
			    	
			    	long lfilebytes = lFileHandle.length();
			    	Log.d(LOGTAG, "getFileLengthMSBnd() 22  filebytes:" + lfilebytes + " fileb - head: " + (lfilebytes - 40) );
			    	lfilebytes = lfilebytes - 40; // FIXME: crappy ballpark guess at headers -- this should be replaced with actually reading wave headers and subtracting from size OR get WAVE data chunk size
			    	long lframeCount = lfilebytes / lframeSize;
			    	
			    	lfileMS = (long) (lframeCount / lfrequencyMS);
			    	Log.d(LOGTAG, "getFileLengthMSBnd() 24  fileMS: " + lfileMS + " frameCount: " + lframeCount + " lfreq: " + lfrequency + " lfreqMS: " + lfrequencyMS);
		    	} catch (Exception e) {
		        	Log.w(LOGTAG, "getFileLengthMSBnd() 55  EXCEPTION e:" + e.getLocalizedMessage());
					e.getLocalizedMessage();
		    	}
    			mFilelen = lfileMS;
    		}
    		
    	} // END if mFilename != null
    	Log.d(LOGTAG, "getFileLengthMSBnd() 99 lfileMS: " + lfileMS);
    	return lfileMS;
    } // END getFileLengthMSBnd()

    
    /**
     * Configure the state change listener, via the binder, allowing an activity to react to state changes.
     * @param newListener with an MPQAudioServiceListener
     */
    public void setListenerBnd(MPQAudioServiceListener newListener) {
 		Log.d(LOGTAG, "setListenerBnd() 00 newListener: " + newListener);
    	mStateListener = newListener;
    }
    	

    
    
    //////////////////////////////////
    //
    // private worker methods for the class, including sub-class MPQAudioThread
    
    /**
     * Show a notification and start running foreground service.
     * @param noticeID int code with an ID for the notification content to display
     * 		    ex: R.string.notification_aservice_rec , defined in strings.xml
	 * @throws IllegalArgumentException if an noticeID is received
     */
    private void showForegroundNotification(int noticeID) {
    	CharSequence ltext;
    	CharSequence ltitle = getText(R.string.notification_aservice);

    	if (R.string.notification_aservice_rec == noticeID) {
    		ltext = getText(R.string.notification_aservice_rec);
    	} else if (R.string.notification_aservice_play == noticeID) {
    		ltext = getText(R.string.notification_aservice_play);
    	} else if (R.string.notification_aservice_pause_rec == noticeID) {
    		ltext = getText(R.string.notification_aservice_pause_rec);
    	} else if (R.string.notification_aservice_pause_play == noticeID) {
    		ltext = getText(R.string.notification_aservice_pause_play);
    	} else if (R.string.notification_aservice_ready == noticeID) {
    		ltext = getText(R.string.notification_aservice_ready);
    	} else {
        	// Raise exception
        	Log.e(LOGTAG, "showNotification() 20 RAISING EXCEPTION");
        	throw new IllegalArgumentException("showNotification() noticeID: " + noticeID);
    	}
    	
        // Set the icon, scrolling text and timestamp
    	NotificationCompat.Builder builder = new NotificationCompat.Builder(this);
    	builder.setSmallIcon(R.drawable.ic_launcher);
    	builder.setContentTitle(ltitle);
    	builder.setContentText(ltext);

    	Intent intent = new Intent(this, MPQMainTabActivity.class);
        // The PendingIntent to launch our activity if the user selects this notification
        PendingIntent lcontentIntent = PendingIntent.getActivity(this, 0, intent, 0); // FIXME: keep or toss this -- PendingIntent.FLAG_CANCEL_CURRENT);
    	builder.setContentIntent(lcontentIntent);

        // Build and send the notification and begin running service in the foreground.
        Notification n = builder.build();
        startForeground(NOTIFICATION_ID, n);
    } // END showForegroundNotification()
    

    /**
     * Request audio focus.  
     */
    private void getAudioFocus() {
 		Log.d(LOGTAG, "getAudioFocus() 00  ");
 		
 		int lrc;
 		lrc = mAM.requestAudioFocus(audioFocusListener, AudioManager.STREAM_MUSIC, AudioManager.AUDIOFOCUS_GAIN);

 		if (lrc == AudioManager.AUDIOFOCUS_REQUEST_GRANTED) {
 			;// nothing for now...
 		} else {
 	 		Log.e(LOGTAG, "getFocus() 30  -- unable to get audio focus.");  // FIXME: do something better. 	ideally a user dialog - but how to trigger from the service -- need error return/state (in addition to current state change)
 		}
 		Log.d(LOGTAG, "getAudioFocus() 99  rc:" + lrc);
    } // END getFocus()
    
    /**
     * Callback listener for audio focus change.
     */
    private AudioManager.OnAudioFocusChangeListener audioFocusListener = new AudioManager.OnAudioFocusChangeListener() {
    	public void onAudioFocusChange(int focusChange) {
    		Log.e(LOGTAG,"onAudioFocusChange() listener called focusChange:" + focusChange);
    		// FIXME: implement response for when we lose focus -- likely pause (play and/or rec)...
            if (focusChange == AudioManager.AUDIOFOCUS_LOSS_TRANSIENT) {
            	// Pause playback
            } else if (focusChange == AudioManager.AUDIOFOCUS_GAIN) {
            	// Resume playback 
            } else if (focusChange == AudioManager.AUDIOFOCUS_LOSS) {
                releaseAudioFocus(); // FIXME: this may not be required, if we trigger the stop action... 
                // Stop playback
            }
    	}
    };
    
    /**
     * Release audio focus.
     * {@link #mAM handle to AudioManager used.
     */
    private void releaseAudioFocus() {
 		Log.d(LOGTAG, "releaseAudioFocus() 00  ");
 		mAM.abandonAudioFocus(audioFocusListener);
    }

    
    /**
     * Trigger UI listeners of state change.
     * {@link #mStateListener} checked if valid, then called.
     */
    private void triggerStateListener() {
        Log.d(LOGTAG, "triggerStateListener() 00 state: " + mState);
    	if (null != mStateListener) {
        	mStateListener.onMPQAudioServiceStateChange(mState);
        }     
    } // END triggerStateListener()
     
    
    /**
     * Trigger UI listeners of audio changes.
     * {@link #mStateListener} checked if valid, then called.
     */
    private void triggerAudioListener() {
        Log.d(LOGTAG, "triggerAudioListener() 00 filename: " + mFilename + " pos: " + mFilepos + " len: " + mFilelen + " vol: " + mVolumeLevel);
        Log.d(LOGTAG, "triggerAudioListener() 10  pos : " + SessionsDB.longSessionMSecString(mFilepos) + " len: " + SessionsDB.longSessionMSecString(mFilelen) );
    	if (null != mStateListener) {
        	mStateListener.onMPQAudioServiceAudioChange(mFilepos, mFilelen, mVolumeLevel);
        }     
    } // END triggerStateListener()
     
    
    

	/**
	 * Prepare and send a message to the child (worker) thread with current state.
	 * Credit: Cristian Baita
	 */
	private void sendCMsgState(int state) {
		Log.d(LOGTAG, "sendCMsgState() 00 state: " + state);        		
		Message messageToChild = new Message();
		Bundle messageData = new Bundle();
//		messageToThread.what = 0;
		messageData.putInt(MPQAudioWorkerThread.ARG_CMSG_STATE, mState);
		messageToChild.setData(messageData);

		// send message to worker thread
		if (null != mWorkerThread) {
			mWorkerThread.getHandler().sendMessage(messageToChild);
		}
		return;
	} // END sendCMsgState()
    

	/**
	 * Prepare and send a message to the child (worker) thread with new cursor position in milliseconds.
	 */
	private void sendCMsgPos(long filepos) {
		Log.d(LOGTAG, "sendCMsgPos() 00 filepos: " + filepos);
		Message messageToChild = new Message();
		Bundle messageData = new Bundle();
////		messageToParent.what = 0;
		messageData.putLong(MPQAudioWorkerThread.ARG_CMSG_FILEPOS, filepos);
		messageToChild.setData(messageData);

		// send message to worker thread
		if (null != mWorkerThread) {
			mWorkerThread.getHandler().sendMessage(messageToChild);
		}
		return;
	} // END sendCMsgPos()
	
} // END class MPQAudioService
