/*
 * MPQGenDialogFragment.java: General dialog boxes for the MPQ app.
 *            Copyright (C) 2012 donfede@casagrau.org
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * created 2012-11-04 donfede@casagrau.org
 *
 * $Id: MPQGenDialogFragment.java,v 1.2 2012-11-05 23:11:14 donfede Exp $
 */

package org.casagrau.mpq.mpq_android;

import android.app.Activity;
import android.app.AlertDialog;
import android.app.Dialog;
import android.content.DialogInterface;
import android.os.Bundle;
import android.support.v4.app.DialogFragment;
import android.util.Log;

/**
 * Flexible dialog constructor.
 * 
 * Dialog that displays an label and two EditText boxes (trying to replicate
 * EditTextPreference), then allows the user to proceed or cancel.  
 * 
 * User feedback is handled by the MPQGenDialogListener interface, which requires two methods for Pos and Neg responses from the user.
 *
 * @param ARG_DIALOG_TYPE int code describing the dialog type.
 * 				ARG_DIALOG_TYPE_SAVE_DISCARD allows the user to save or discard edits
 * @param ARG_DIALOG_TITLE String to display as a title in the dialog box.  
 * @param ARG_DIALOG_ID int to identify the dialog box in the callbacks.  Negative values represent unknown/invalid state.
 */
public class MPQGenDialogFragment extends DialogFragment {
	private static final String LOGTAG = "MPQGenDialogFragment";

	public static final String ARG_DIALOG_TYPE = "mpqgen_dialog_type"; // Bundle handle to dialog type code
	public static final int ARG_DIALOG_TYPE_SAVE_DISCARD = 0;
	public static final int ARG_DIALOG_TYPE_ADD = 1;
	public static final String ARG_DIALOG_TITLE = "mpqgen_dialog_title"; // Bundle handle to dialog title String
	public static final String ARG_DIALOG_ID = "mpqgen_dialog_id"; // Bundle handle to dialog id 
	
	private String mDialogTitle;
	private int mDialogId = -1; // Identifier for the dialog.  Negative values represent unknown/invalid state... probably should through an exception [This is used as there seems to be no easy way to distinguish between who called a dialog/what role a generic dialog was for].   
	
	//////////////
	// Callback listener configuration
	/**
	 * Define callback interface to respond to dialog "yes/no" clicks. [a bit of
	 * a pain for a simple dialog...]
	 */
	public interface MPQGenDialogListener {
		public void onMPQGenDialogPositiveClick(DialogFragment dialog, int id);

		public void onMPQGenDialogNegativeClick(DialogFragment dialog, int id);
	}

	private MPQGenDialogListener mListener;	// Handle back to the Listener.


	@Override
	public void onAttach(Activity activity) {
		super.onAttach(activity);
		Log.d(LOGTAG, "onAttach() 00");
		
		mListener = (MPQGenDialogListener) getTargetFragment();
		Log.d(LOGTAG, "onAttach() 10 mListener from fragment:" + mListener);
		if (null == mListener) {
			mListener = (MPQGenDialogListener) activity;
			Log.d(LOGTAG, "onAttach() 10 mListener from activity:" + mListener);
		}
		if (null == mListener) {
			Log.d(LOGTAG, "onAttach() Exception - no related listener");
			throw new ClassCastException(activity.toString() + " must implement MPQGenDialogListener");
		}
	} // END onAttach()

	
	
	/**
	 * Create dialog box corresponding with parameter input.
	 */
	@Override
	public Dialog onCreateDialog(Bundle savedInstanceState) {
		// Be sure to call the super class.
		super.onCreate(savedInstanceState);

		Log.d(LOGTAG, "DEBUG onCreateDialog() 00");
		Log.d(LOGTAG, "DEBUG onCreateDialog() 10 bundle:" + savedInstanceState);
		Log.d(LOGTAG, "DEBUG onCreateDialog() 12 mDialogTitle:" + mDialogTitle);

		// Process arguments. FIXME: review, probably trash this section
		//        if (savedInstanceState != null) { 
		//        	Log.d(LOGTAG, "DEBUG onCreateDialog() 20");
		//        	mDialogTitle = savedInstanceState.getString(ARG_DIALOG_TITLE);
		//        }
		int ltype = getArguments().getInt(ARG_DIALOG_TYPE); 
		mDialogTitle = getArguments().getString(ARG_DIALOG_TITLE);
		Log.d(LOGTAG, "DEBUG onCreateDialog() 22 mDialogTitle:" + mDialogTitle);
		mDialogId = getArguments().getInt(ARG_DIALOG_ID, -1);
		Log.d(LOGTAG, "DEBUG onCreateDialog() 24 mDialogId:" + mDialogId);

		AlertDialog.Builder builder = new AlertDialog.Builder(getActivity());
		builder.setTitle(mDialogTitle);
		int lresPos = android.R.string.ok ; // Positive button resource.
		int lresNeg = android.R.string.cancel; // Negative button resource.
		if (ARG_DIALOG_TYPE_SAVE_DISCARD == ltype) {
			lresPos = R.string.dialog_mpqgen_save;
			lresNeg = R.string.dialog_mpqgen_discard;
			//builder.setMessage(R.string.foostring);
			//builder.setView(  getActivity().getLayoutInflater().inflate(R.layout.fragment_dialog_mpqadd, null));						
		} else if (ARG_DIALOG_TYPE_ADD == ltype) {
			lresPos = R.string.dialog_mpqadd_button_pos;
			lresNeg = R.string.dialog_mpqadd_button_neg;
			builder.setView(  getActivity().getLayoutInflater().inflate(R.layout.fragment_dialog_mpqadd, null));			
		}
		builder.setPositiveButton(lresPos, 
				new DialogInterface.OnClickListener() {
			public void onClick(DialogInterface dialog, int id) {
				// Send the positive button event back to the host activity
				mListener.onMPQGenDialogPositiveClick(MPQGenDialogFragment.this, mDialogId);
			} // END onClick()
		});
		builder.setNegativeButton(lresNeg,
				new DialogInterface.OnClickListener() {
			public void onClick(DialogInterface dialog, int which) {
				mListener.onMPQGenDialogNegativeClick(MPQGenDialogFragment.this, mDialogId);
			} // END onClick()
		});
		return builder.create();
	} // END onCreateDialog
	

// FIXME: review, probably trash this section	
//    /** 
//     * Save state of the app - mCurrentTab only for now.
//     * 
//     */
//    @Override
//    public void onSaveInstanceState(Bundle savedInstanceState) {
//    	savedInstanceState.putString(ARG_DIALOG_TITLE, mDialogTitle);
//    	Log.d(LOGTAG, "DEBUG onSaveInstanceState()");
//    	// Always call the superclass so it can save the view hierarchy state
//    	super.onSaveInstanceState(savedInstanceState);
//    }

}
